////////////////////////////////////////////////////////////////////// 
// PreMeta.cpp 
// (c) 2012-2013 Shuang Feng, Dajiang Liu, Goncalo Abecasis
// 
// This file is distributed as part of the RareMetalWorker source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile RareMetalWorker.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Wednesday November 28, 2012
// 

#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include "PreMeta.h"
#include "snpHWE.h"
#include "WritePDF.h"
#define MATHLIB_STANDALONE
#include <Rmath.h>
#include "InputFile.h"

String PreMeta::vcfInput = ""; //input annotated vcf file name here
int PreMeta::Xstart = 2699520;
int PreMeta::Xend = 154931044;
bool PreMeta::dosage =false;
bool PreMeta::stdErr =true;
bool PreMeta::genoFromPed =false;
bool PreMeta::genoFromVCF =false;
bool PreMeta::FounderFreq =false;
bool PreMeta::zipOutput=false;
String PreMeta::outputFile = "";
String PreMeta::xLabel = "X";
int PreMeta::window = 1000000;
int PreMeta::maleLabel = 1;
int PreMeta::femaleLabel =2;
bool PreMeta::correctGC =false;

PreMeta::PreMeta(){}

PreMeta::~PreMeta()
{}

void PreMeta::VCFSanityCheck()
{
   reader.open(vcfInput,header);
   String chrom;
   int position;
   while(reader.readRecord(record))
   {
      chrom = record.getChromStr();
      position = record.get1BasedPosition();
      break;
   }
   reader.close();

   if(chrom.SlowFind("chr",0)!=-1)
   {
      //chromosome in VCF has "chr" in it
      reader.open(vcfInput,header);
      String prev_chrom;
      int prev_pos;
      while(reader.readRecord(record))
      {
	 prev_chrom = chrom;
	 prev_pos = position;
	 chrom = record.getChromStr();
	 position = record.get1BasedPosition();
	 //compare here
	 if(chrom==prev_chrom)
	 {
	    if(position <= prev_pos)
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	 }
	 else 
	 {
	    prev_chrom = prev_chrom.SubStr(3);
	    chrom = chrom.SubStr(3);
	    if(atoi(chrom.c_str()) < atoi(prev_chrom.c_str()))
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	    else
	    {
	       continue;
	    }
	 }
	 reader.close();
      }
   }
   else 
   {
      reader.open(vcfInput,header);
      String prev_chrom;
      int prev_pos;
      while(reader.readRecord(record))
      {
	 prev_chrom = chrom;
	 prev_pos = position;
	 chrom = record.getChromStr();
	 position = record.get1BasedPosition();
	 //compare here
	 if(chrom==prev_chrom)
	 {
	    if(position <= prev_pos)
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	 }
	 else if(atoi(chrom.c_str()) < atoi(prev_chrom.c_str())) 
	 {
	    error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	    break;
	 }
	 else 
	 {
	    continue;
	 }
      }
      reader.close();
   }
}

double PreMeta::CalculateCov(FastFit & engine,FastTransform & trans, Pedigree & ped, Matrix & genotypeAll,Vector & sigma2,int m)
{
   double cov = 0.0;
   if(FastFit::unrelated)
   {    
      for(int p=0;p<trans.persons;p++)
      {     
	 cov += genotypeAll[p][0] * genotypeAll[p][m];
      }     
      cov *= engine.sigma2Hat;
   }    
   else 
   {    
      for(int p=0;p<trans.persons;p++)
      {     
	 cov += genotypeAll[p][0] * genotypeAll[p][m]/sigma2[p];
      }     
   }   
   //printf("cov is: %g\n",cov); 
   //cov /= trans.persons;
   return cov;
}

void PreMeta::Run(IFILE SCOREoutput,IFILE SCOREcov,Pedigree & ped, FastTransform & trans,FastFit & engine,GroupFromAnnotation & group, FILE * log)
{
   Vector chisq_before_GCcorrect,maf_GC;
   double error = 0.0;
   StringArray chr_plot;
   Vector pos_plot;
   printf("  Calculating association test statistics ...");
   fprintf(log,"  Calculating association test statistics ...");
   double averageAF=0.0;
   int marker_count=0;
   /*
      if(genoFromVCF)
      {
   //This is to make sure VCF file is sorted
   VCFSanityCheck();
   }
    */
   PrintHeader(SCOREoutput,engine,trans,ped);
   ifprintf(SCOREcov,"##ProgramName=RareMetalWorker\n");
   ifprintf(SCOREcov,"##Version=%s\n",VERSION);
   ifprintf(SCOREcov,"#CHROM\tCURRENT_POS\tMARKERS_IN_WINDOW\tCOV_MATRICES\n");

   Matrix genotypeAll; //this matrix has the all marker genotype within a window
   genotypeAll.Dimension(0,0);

   IntArray markerInLD;

   Vector sigma2,transGeno,pvalueAll,pvalue5,pvalue1;
   transGeno.Dimension(trans.persons);
   sigma2.Dimension(trans.persons);

   if(genoFromPed)
   {
      //TODO: this assumes all markers are sorted by chromosome and position
      StringArray name;
      String initial_chr;
      int initial_pos;
      int startingMarker;
      for(int i=0;i<ped.markerCount;i++)
      {
	 name.AddTokens(ped.markerNames[i],":");
	 /*
	    if(name.Length()<=1)
	    error("ERROR! Marker names in DAT file has to be in CHR:POS format.\n");
	  */
	 if(name.Length()<=1)
	 {
	    printf("WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from analysis.\n",ped.markerNames[i].c_str());
	    fprintf(log,"WARNING: marker name %s in DAT file is not in CHR:POS format and will be skipped from analysis.\n",ped.markerNames[i].c_str());
	    continue;
	 }
	 initial_chr = name[0];
	 initial_pos = atoi(name[1].c_str());
	 startingMarker=i;
	 break;
      }

      for(int m=startingMarker;m<ped.markerCount;m++)
      {
	 int status = GetGenotypeVectorPED(trans,ped,m,log);
	 if(status==2)
	    continue;

	 if(status==1)
	 {
	    if(stdErr)
	       ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\tNA\tNA\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2);
	    else
	       ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\tNA\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2);
	    continue;
	 }
	 averageAF += markerMaf;
	 marker_count++;
	 //output score statistics
	 if(FastFit::unrelated)
	 {
	    double numerator = 0.0;
	    double denominator = 0.0;
	    double chisq,pvalue,effSize;
	    for(int i=0;i<trans.persons;i++)
	    {
	       numerator += engine.residuals_unrelated[i]*genotype[i];
	       denominator += genotype[i]*genotype[i];
	    }
	    if(denominator==0)
	       effSize = _NAN_;
	    else
	       effSize = numerator/denominator;

	    denominator *= engine.sigma2Hat;
	    if(denominator==0) 
	       pvalue = _NAN_;
	    else 
	    {
	       chisq = numerator*numerator/denominator;
	       chisq_before_GCcorrect.Push(chisq);
	       pvalue = pchisq(chisq, 1,0,0);
	       if(FounderFreq)
	       {
		  if(founderMaf<0.01)
		     pvalue1.Push(pvalue);
		  if(founderMaf<0.05)
		     pvalue5.Push(pvalue);
		  maf_GC.Push(founderMaf);
	       }
	       else
	       {
		  if(markerMaf<0.01)
		     pvalue1.Push(pvalue);
		  if(markerMaf<0.05)
		     pvalue5.Push(pvalue);
		  maf_GC.Push(markerMaf);
	       }
	       pvalueAll.Push(pvalue);
	       chr_plot.Push(chr);
	       pos_plot.Push(pos);
	       //calculate standard error here
	       if(stdErr)
	       {
		  Vector sigma2;
		  error = CalculateStdErr(engine,trans,genotype,sigma2,denominator);
	       }
	    }
	    if(pvalue==_NAN_)
	    {
	       if(stdErr)
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	       else
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	    }
	    else
	    {
	       chisqred.Push(chisq);
	       if(stdErr)
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue,error);
	       else
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue);
	    }
	 }
	 else
	 {
	    //get transGeno
	    if(trans.pedKin)
	    {
	       transGeno.Zero();
	       int index=0;
	       for (int f = 0; f < ped.familyCount; f++)
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			}
			index++;
		     }
		  }
	    } else transGeno.Product(trans.transU,genotype);

	    //calculate score statistics and output here
	    double sigma_g2 = engine.sigma_g2Hat;
	    double sigma_e2 = engine.sigma_e2Hat;

	    if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	    {
	       for(int i=0;i<trans.persons;i++)
		  sigma2[i] = trans.D[i];
	    }
	    else
	    {
	       for(int i=0;i<trans.persons;i++)
	       {
		  sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
	       }
	    }
	    double numerator = 0.0;
	    double denominator = 0.0;
	    double fixedEffect,chisq,pvalue,effSize;
	    for(int i=0;i<trans.persons;i++)
	    {
	       fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
	       numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
	       denominator += transGeno[i]*transGeno[i]/sigma2[i];
	    }
	    if(denominator==0.0)
	    {
	       pvalue=_NAN_;
	       effSize = _NAN_;
	    }
	    else
	    {
	       chisq = numerator*numerator/denominator;
	       chisq_before_GCcorrect.Push(chisq);
	       pvalue = pchisq(chisq, 1,0,0);
	       effSize = numerator/denominator;
 if(FounderFreq)
               {
                  if(founderMaf<0.01)
                     pvalue1.Push(pvalue);
                  if(founderMaf<0.05)
                     pvalue5.Push(pvalue);
                  maf_GC.Push(founderMaf);
               }
               else
               {
                  if(markerMaf<0.01)
                     pvalue1.Push(pvalue);
                  if(markerMaf<0.05)
                     pvalue5.Push(pvalue);
                  maf_GC.Push(markerMaf);
               }
	       pvalueAll.Push(pvalue);
	       chr_plot.Push(chr);
	       pos_plot.Push(pos);

	       //calculate standard error here
	       if(stdErr)
		  error = CalculateStdErr(engine,trans,transGeno,sigma2,denominator);
	    }
	    if(pvalue==_NAN_)
	       if(stdErr)
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	       else
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	    else
	    {
	       chisqred.Push(chisq);
	       if(stdErr)
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue,error);
	       else
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue);
	    }
	 }

	 if(chr != initial_chr)
	 {
	    //output cov for the rest of the markers saved in genotypeAll
	    int markers = genotypeAll.cols;
	    for(int i=0;i<markers;i++)
	    {
	       initial_pos = markerInLD[0];
	       //       printf("current position: %d; initial position: %d,markerInLD length %d\t",pos,initial_pos,markerInLD.Length());
	       ifprintf(SCOREcov,"%s\t%d\t",initial_chr.c_str(),initial_pos);
	       for(int j=0;j<markerInLD.Length();j++)
	       {
		  ifprintf(SCOREcov,"%d,",markerInLD[j]);
	       }
	       ifprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  ifprintf(SCOREcov,"%g,",cov);
	       }
	       ifprintf(SCOREcov,"\n");

	       //reorganize the trackers
	       if(markerInLD.Length()>0)
	       {
		  markerInLD.Delete(0);
		  genotypeAll.DeleteColumn(0);
	       }
	       //    printf("markerInID length after %d\n",markerInLD.Length());
	    }

	    //Start a new sequence
	    initial_chr = chr;
	    initial_pos = pos;
	    markerInLD.Clear();
	    genotypeAll.Dimension(0,0);
	 }
	 else
	 {
	    while(pos - initial_pos > window && markerInLD.Length() >0 )
	    {
	       //output cov for marker at initial_pos
	       //      printf("current position: %d; initial position: %d,markerInLD length %d\t",pos,initial_pos,markerInLD.Length());
	       ifprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	       for(int i=0;i<markerInLD.Length();i++)
	       {
		  ifprintf(SCOREcov,"%d,",markerInLD[i]);
	       }
	       ifprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  ifprintf(SCOREcov,"%g,",cov);
	       }
	       ifprintf(SCOREcov,"\n");

	       markerInLD.Delete(0);
	       genotypeAll.DeleteColumn(0);
	       initial_pos = markerInLD[0];
	    }
	    //	    printf("markerInID length after %d\n",markerInLD.Length());
	 }
	 //if not, then update current sequence  
	 markerInLD.Push(pos);
	 initial_pos = markerInLD[0];
	 genotypeAll.GrowTo(trans.persons,genotypeAll.cols+1,0.0);
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(FastFit::unrelated)
	       genotypeAll[i][genotypeAll.cols-1] = genotype[i];
	    else
	       genotypeAll[i][genotypeAll.cols-1] = transGeno[i];
	 }
      }
      //output cov for the rest of the markers saved in genotypeAll
      int markers = genotypeAll.cols;
      for(int i=0;i<markers;i++)
      {
	 initial_pos = markerInLD[0];
	 ifprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	 for(int j=0;j<markerInLD.Length();j++)
	 {
	    ifprintf(SCOREcov,"%d,",markerInLD[j]);
	 }
	 ifprintf(SCOREcov,"\t");
	 for(int m=0;m<genotypeAll.cols;m++)
	 {
	    double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
	    ifprintf(SCOREcov,"%g,",cov);
	 }
	 ifprintf(SCOREcov,"\n");

	 //reorganize the trackers
	 if(markerInLD.Length()>0)
	 {
	    markerInLD.Delete(0);
	    genotypeAll.DeleteColumn(0);
	 }
      }
   }

   if(genoFromVCF)
   {
      StringArray chromosome;
      reader.open(vcfInput,header);
      reader.readVcfIndex();
      const Tabix* indexPtr = reader.getVcfIndex();
      for(int i = 0; i < indexPtr->getNumRefs(); i++)
      {
	 chromosome.Push(indexPtr->getRefName(i));
      }
      reader.close();

      //loop through chromosomes by reading tabix data
      for(int i=0;i<chromosome.Length();i++)
      {
	 reader.open(vcfInput,header);
	 reader.readVcfIndex();
	 reader.setReadSection(chromosome[i].c_str());
	 int initial_pos = 0;

	 while(reader.readRecord(record))
	 {
	    bool status;
	    if(dosage)
	       status = GetGenotypeVectorVCFDosage(trans,ped);
	    else
	       status = GetGenotypeVectorVCF(trans,ped);
	    if(!status)
	    {
	       if(stdErr)
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\tNA\tNA\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2);
	       else
		  ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\tNA\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2);
	       continue;
	    }
	    marker_count++;
	    averageAF += markerMaf;

	    if(marker_count==1) 
	       initial_pos = pos;

	    //Output score statistics here
	    if(FastFit::unrelated)
	    {
	       double numerator = 0.0;
	       double denominator = 0.0;
	       double chisq,pvalue,effSize;
	       for(int i=0;i<trans.persons;i++)
	       {
		  //fixedEffect=trans.X[i].InnerProduct(engine.betaHat);
		  numerator += engine.residuals_unrelated[i]*genotype[i];
		  denominator += genotype[i]*genotype[i];
	       }

	       if(denominator==0)
		  effSize = _NAN_;
	       else
		  effSize = numerator/denominator;

	       denominator *= engine.sigma2Hat;
	       if(denominator==0)
		  pvalue = _NAN_;
	       else 
	       {
		  chisq = numerator*numerator/denominator;
		  chisq_before_GCcorrect.Push(chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  if(FounderFreq)
		  {
		     if(founderMaf<0.01)
			pvalue1.Push(pvalue);
		     if(founderMaf<0.05)
			pvalue5.Push(pvalue);
		     maf_GC.Push(founderMaf);
		  }
		  else
		  {
		     if(markerMaf<0.01)
			pvalue1.Push(pvalue);
		     if(markerMaf<0.05)
			pvalue5.Push(pvalue);
		     maf_GC.Push(markerMaf);
		  }
		  pvalueAll.Push(pvalue);
		  chr_plot.Push(chr);
		  pos_plot.Push(pos);
		  //calculate standard error here
		  if(PreMeta::stdErr)
		  {
		     Vector sigma2;
		     error = CalculateStdErr(engine,trans,genotype,sigma2,denominator);
		  }
	       }
	       if(pvalue==_NAN_)
		  if(stdErr)
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
		  else
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	       else
	       {
		  chisqred.Push(chisq);
		  if(stdErr)
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue,error);
		  else
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue);
	       }
	    }
	    else
	    {
	       //get transGeno
	       if(!trans.empKin)
	       {
		  transGeno.Zero();
		  int index=0;
		  for (int f = 0; f < ped.familyCount; f++)
		     if (trans.pPheno[f].Length() > 0)
		     {
			int count = trans.pPheno[f].Length();
			int block = index;
			for(int i=0;i<count;i++)
			{
			   for(int c=0;c<count;c++)
			   {
			      transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			   }
			   index++;
			}
		     }
	       } else transGeno.Product(trans.transU,genotype);

	       //calculate score statistics and output here
	       double sigma_g2 = engine.sigma_g2Hat;
	       double sigma_e2 = engine.sigma_e2Hat;
	       // printf("%g %g\n",sigma_g2,sigma_e2);
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	       double numerator = 0.0;
	       double denominator = 0.0;
	       double fixedEffect,chisq,pvalue,effSize;
	       for(int i=0;i<trans.persons;i++)
	       {
		  fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
		  numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
		  denominator += transGeno[i]*transGeno[i]/sigma2[i];
	       }
	       if(denominator==0.0) 
	       {
		  pvalue=_NAN_;
		  effSize = _NAN_;
	       }
	       else 
	       {
		  chisq = numerator*numerator/denominator;
		  chisq_before_GCcorrect.Push(chisq);
		  pvalue = pchisq(chisq, 1,0,0);
		  effSize = numerator/denominator;
		  if(FounderFreq)
		  {
		     if(founderMaf<0.01)
			pvalue1.Push(pvalue);
		     if(founderMaf<0.05)
			pvalue5.Push(pvalue);
		     maf_GC.Push(founderMaf);
		  }
		  else
		  {
		     if(markerMaf<0.01)
			pvalue1.Push(pvalue);
		     if(markerMaf<0.05)
			pvalue5.Push(pvalue);
		     maf_GC.Push(markerMaf);
		  }
		  pvalueAll.Push(pvalue);
		  chr_plot.Push(chr);
		  pos_plot.Push(pos);

		  //calculate standard error here
		  if(PreMeta::stdErr)
		     error = CalculateStdErr(engine,trans,transGeno,sigma2,denominator);
	       }
	       if(pvalue==_NAN_)
		  if(stdErr)
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
		  else
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\tNA\tNA\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator));
	       else
	       {
		  chisqred.Push(chisq);
		  if(stdErr)
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue,error);
		  else
		     ifprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%d\t%d\t%d\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,callRate,hwe_pvalue,hom1,het,hom2,numerator,sqrt(denominator),effSize,pvalue);
	       }
	    }

	    //check if position has passed the window
	    while(pos - initial_pos > window && markerInLD.Length() >0 )
	    {
	       //output cov for marker at initial_pos
	       ifprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	       for(int i=0;i<markerInLD.Length();i++)
	       {
		  ifprintf(SCOREcov,"%d,",markerInLD[i]);
	       }
	       ifprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  ifprintf(SCOREcov,"%g,",cov);
	       }
	       ifprintf(SCOREcov,"\n");

	       //reorganize the trackers
	       markerInLD.Delete(0);
	       initial_pos = markerInLD[0];
	       genotypeAll.DeleteColumn(0);
	    }

	    //push the new marker in
	    markerInLD.Push(pos);
	    initial_pos = markerInLD[0];
	    genotypeAll.GrowTo(trans.persons,genotypeAll.cols+1,0.0);
	    for(int i=0;i<trans.persons;i++)
	    {
	       if(FastFit::unrelated)
		  genotypeAll[i][genotypeAll.cols-1] = genotype[i];
	       else
		  genotypeAll[i][genotypeAll.cols-1] = transGeno[i];
	    }
	 }
	 //output the LD matrix for the rest of the markers in genotypeAll
	 int markers = genotypeAll.cols;
	 for(int i=0;i<markers;i++)
	 {
	    ifprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	    for(int j=0;j<markerInLD.Length();j++)
	    {
	       ifprintf(SCOREcov,"%d,",markerInLD[j]);
	    }
	    ifprintf(SCOREcov,"\t");
	    for(int m=0;m<genotypeAll.cols;m++)
	    {
	       double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
	       ifprintf(SCOREcov,"%g,",cov);
	    }
	    ifprintf(SCOREcov,"\n");

	    //reorganize the trackers
	    markerInLD.Delete(0);
	    initial_pos = markerInLD[0];
	    genotypeAll.DeleteColumn(0);
	 }
	 reader.close();
      }
   }
   if(averageAF/marker_count>0.5)
   {
      printf("\n\nWarning: Please check for allele flips, the reference allele frequency in your sample is <0.5 on average!\n\n");
      fprintf(log,"\n\nWarning: Please check for allele flips, the reference allele frequency in your sample is <0.5 on average!\n\n");
   }
   if(chisqred.Length()>0)
   {
      chisqred.Sort();
      //lambda = chisqred[0.5]/0.4549364;
   }

   PDF pdf;
   WritePDF writepdf;
   String filename;

   if(outputFile != "")
   {
      if(outputFile.Last()=='.' ||outputFile.Last()=='/')
	 filename = outputFile + FastFit::traitName + ".plots.pdf";
      else
	 filename = outputFile + "." + FastFit::traitName + ".plots.pdf";
   }
   else
      filename = FastFit::traitName + ".plots.pdf";

   //get labels of hits
   StringArray geneLabel;
   if(group.labelHits)
   {
      String target_chr="";
      int target_pos=0;
      int target=0;
      double target_pvalue = _NAN_;
      for(int lb=0;lb<pvalueAll.Length();lb++)
      {
	 if(pvalueAll[lb] < 0.05/pvalueAll.Length())
	 {
	    //if this locus is annotated already
	    if(target_chr==chr_plot[lb] && target_pos>pos_plot[lb]-1000000)
	    {
	       //if this is the higher peak then update the annotation of this locus
	       if(pvalueAll[lb]<target_pvalue)
	       {
		  String current_anno = group.AnnotateSingleVar(chr_plot[lb],pos_plot[lb]);
		  int distance = 1000;
		  while(current_anno=="" && distance<=1000000)
		  {
		     current_anno = group.AnnotateSingleVar(chr_plot[lb],pos_plot[lb]+distance);
		     if(current_anno=="")
			current_anno = group.AnnotateSingleVar(chr_plot[lb],pos_plot[lb]-distance);
		     distance += 1000;
		  }
		  if(geneLabel[target]!="" && current_anno!="" && geneLabel[target].Find(current_anno)==-1)
		     current_anno = geneLabel[target] + "/" + current_anno;
		  if(geneLabel[target]!="" && current_anno=="")
		     current_anno = geneLabel[target];

		  geneLabel.Push(current_anno);
		  geneLabel[target]="";
		  target = geneLabel.Length()-1;
		  target_pvalue = pvalueAll[lb];
		  target_pos = pos_plot[lb];
		  target_chr = chr_plot[lb];
	       }
	       //otherwise, leave the original annotation of this locus
	       else
		  geneLabel.Push("");
	    }
	    else
	    {
	       geneLabel.Push(group.AnnotateSingleVar(chr_plot[lb],pos_plot[lb]));
	       target_chr = chr_plot[lb];
	       target_pos = pos_plot[lb];
	       target_pvalue = pvalueAll[lb];
	       target = geneLabel.Length()-1;
	    }
	 }
	 else
	    geneLabel.Push("");
      }
   }

   //calculate GC
   Vector tmp;
   tmp.Copy(chisq_before_GCcorrect);
   tmp.Sort();
double GC=_NAN_;
if(tmp.Length()>0)
   double GC = tmp[0.5]/0.456;
else
printf("\n\n Warning: GC is not valid. One possible reason is that all of your variants are monomorphic.\n");

   pdf.OpenFile(filename);
   String title = FastFit::traitName;
   String demo = "GC = ";
   demo+= GC;
if(tmp.Length()>0)
{
   printf("\n    Genomic Control is: %g\n",GC);
   ifprintf(SCOREoutput,"#Genomic control is: %g\n",GC);
 }
  if(group.labelHits)
      writepdf.Draw(pdf,geneLabel,pvalueAll,pvalue1,pvalue5,chr_plot,pos_plot,title,demo,true);
   else
      writepdf.Draw(pdf,geneLabel,pvalueAll,pvalue1,pvalue5,chr_plot,pos_plot,title,demo,false);

   if(GC>1.05 || GC<0.95 ||correctGC)
   {
      Vector pvalue_GCcorrected,pvalue1_GCcorrected,pvalue5_GCcorrected;
      chisq_before_GCcorrect.Multiply(1.0/GC);
      for(int i=0;i<chisq_before_GCcorrect.Length();i++)
      {
	 double p = pchisq(chisq_before_GCcorrect[i],1,0,0);
	 pvalue_GCcorrected.Push(p);
	 if(maf_GC[i]<0.01)
	    pvalue1_GCcorrected.Push(p);
	 if(maf_GC[i]<0.05)
	    pvalue5_GCcorrected.Push(p);
      }
      title += " (GC corrected)";
      if(group.labelHits)
	 writepdf.Draw(pdf,geneLabel,pvalue_GCcorrected,pvalue1_GCcorrected,pvalue5_GCcorrected,chr_plot,pos_plot,title,demo,true);
      else
	 writepdf.Draw(pdf,geneLabel,pvalue_GCcorrected,pvalue1_GCcorrected,pvalue5_GCcorrected,chr_plot,pos_plot,title,demo,false);
   }
   pdf.CloseFile();
}

double PreMeta::CalculateStdErr(FastFit & engine,FastTransform & trans,Vector & geno, Vector & sigma2,double denominator)
{
   double err;
   if(FastFit::unrelated)
   {
      if(FastFit::makeResiduals)
      {
	 double b = geno.Sum();
	 err = engine.sigma2Hat*trans.persons/(trans.persons*denominator-b*b);
      }
      else
      {
	 Matrix transX;
	 transX.Transpose(trans.X);
	 Vector c;
	 c.Dimension(trans.X.cols);
	 for(int i=0;i<trans.X.cols;i++)
	    c[i] = geno.InnerProduct(transX[i]);
	 Vector tmp;
	 tmp.Dimension(trans.X.cols);
	 for(int i=0;i<trans.X.cols;i++)
	    tmp[i] = c.InnerProduct(trans.inv[i]);
	 err =  engine.sigma2Hat/(denominator - tmp.InnerProduct(c));
      }
   }
   else
   {
      Vector UDG;
      UDG.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 UDG[i] = geno[i]/sqrt(sigma2[i]);

      if(FastFit::makeResiduals)
      {
	 double b = UDG.Sum();
	 err = 1.0*trans.persons/(trans.persons*denominator - b*b); 
      }
      else
      {
	 Matrix transX;
	 transX.Transpose(trans.UDX);
	 Vector c;
	 c.Dimension(trans.UDX.cols);
	 for(int i=0;i<trans.UDX.cols;i++)
	    c[i] = geno.InnerProduct(transX[i]);
	 Vector tmp;
	 tmp.Dimension(trans.X.cols);
	 for(int i=0;i<trans.X.cols;i++)
	    tmp[i] = c.InnerProduct(trans.inv[i]);
	 err = 1.0/(denominator - tmp.InnerProduct(c));
      }
   }

   err = sqrt(err);
   return err;
}

bool PreMeta::GetGenotypeVectorVCFDosage(FastTransform & trans, Pedigree & ped)
{
   bool status = false;

   founderMaf = markerMaf = 0.0;
   rareAllele = record.getAltStr();
   refAllele = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   if(refAllele=="."||rareAllele==".")
      return status;

   int n=0,nf=0;
   double fmaf = 0.0, maf=0.0;
   founderMaf = markerMaf = 0.0;
   VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

   bool XStatus = false;
   if(chr==xLabel && pos>=Xstart && pos<=Xend)
      XStatus=true;

   //printf("Now working on marker %d.\n",markerIdx);
   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   mac=0.0;

   //Loop through samples according to trans.pPheno 
   //read in dosage into genotype vector
   int idx=0,nmiss=0;
   double mean = 0.0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid;

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s==-1)
	    error("ERROR! Check sample from VCF and PED file.\n",sample.c_str());
	 if(s!=-1)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(strcmp((*geno).c_str(),".")>0)
	    {
	       float dose = atof((*geno).c_str());
	       genotype[idx] = dose;
	       if(XStatus && ped[trans.samplePEDIDHash.Integer(sample)].sex==maleLabel && dose >0)
	       {
		  genotype[idx] = 2;
	       }
	       n++;
	       maf += genotype[idx];
	       int founder = trans.foundersHash.Integer(sample);
	       if(founder!=-1)
	       {
		  fmaf+=dose;
		  nf++;
	       }
	    }
	    else
	    {
	       nmiss++;
	    }
	 }
	 idx++;
      }
   }
   mac = maf;
   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0.0)
      fmaf = 1.0/(2.0*nf);
   founderMaf = fmaf;
   markerMaf = maf;

   mean = 2.0 * markerMaf;
   if(maf==0.0 || 1.0-maf==0.0)
      return status;

   status = true;
   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
	 genotype[i] = 0.0;
      else
	 genotype[i] -= mean;
   }
   callRate = 1.0-(double) nmiss/trans.persons;
   hwe_pvalue = 1.0;
   het=0; hom1=0; hom2=0;

   return status;
}

bool PreMeta::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped)
{
   bool status = false;

   founderMaf = markerMaf = 0.0;
   rareAllele = record.getAltStr();
   refAllele = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   //Loop through samples according to trans.pPheno 
   het=0;hom1=0;hom2=0;
   int numMiss =0;
   double mean =0.0;
   int n=0,nf=0;
   double fmaf=0.0,maf=0.0;
   mac=0.0;

   /*
      if(refAllele=="."||rareAllele==".")
      return status;
    */
   bool XStatus = false;
   if(chr==xLabel && pos>=Xstart && pos<=Xend)
      XStatus=true;

   //CalculateCov and output score statistic here
   //   printf("Now working on marker %d.\n",markerIdx);
   genotype.Dimension(trans.persons);
   genotype.Zero();

   int idx=0;
   //printf("raw genotypes are:\n");
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid.c_str();
	 bool miss_stat = false;

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s==-1)
	    error("ERROR! Check sample %s in VCF and PED file.\n",sample.c_str());
	 int numGTs = record.getNumGTs(s);
	 int sum = 0;
	 for(int j = 0; j < numGTs; j++)
	 {
	    int a = record.getGT(s,j); 
	    if(a==VcfGenotypeSample::MISSING_GT)
	    {
	       numMiss++;
	       genotype[idx] = -1.0;
	       //printf("%s\t%g\n",sample.c_str(),genotype[idx]);
	       idx++;
	       miss_stat = true;
	       break;
	    }
	    sum += a;
	 }
	 if(miss_stat)
	 {
	    continue;
	 }
	 int founder = trans.foundersHash.Integer(sample);
	 if(XStatus && sum==1 && ped[trans.samplePEDIDHash.Integer(sample)].sex==maleLabel)
	 {
	    genotype[idx] = 2.0;
	    hom2++;
	 }
	 else
	 {
	    genotype[idx] = sum;
	    if(sum==1)
	       het++;
	    else if(sum==0)
	       hom1++;
	    else if(sum==2)
	       hom2++;
	 }
	 maf += genotype[idx];
	 n++;

	 if(founder!=-1)
	 {
	    nf++;
	    fmaf+=genotype[idx];
	 }
	 idx++;
      }
   }

   mac = maf;
   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0.0)
      fmaf = 1.0/(2.0*nf);

   callRate = 1.0-(double) numMiss/trans.persons;

   hwe_pvalue = 0.0;
   //Get hwe_pvalue using the entire sample
   if(het==0 && hom1==0 && hom2==0)
   {
      //printf("marker %s:%d has problem.\n",chr.c_str(),pos);
      refAllele = rareAllele = ".";
      return status;
   }

   hwe_pvalue = SNPHWE(het,hom1,hom2);

   founderMaf = fmaf;
   markerMaf = maf;

   if(maf==0.0 || 1.0-maf==0.0)
      return status;

   status = true;

   /*
      if(markerMaf > 0.5)
      {
      String tmp;
      tmp=refAllele;
      refAllele = rareAllele;
      rareAllele = tmp;
      markerMaf = 1.0-markerMaf;
      founderMaf = 1.0-founderMaf;
      for(int g=0;g<trans.persons;g++)
      {
      if(genotype[g]!=-1.0)
      genotype[g] = 2.0 -genotype[g];
      }
      int tmp_hom2 = hom2;
      hom2 = hom1;
      hom1 = tmp_hom2;
      hwe_pvalue = SNPHWE(het,hom1,hom2);
      }
    */
   mean = 2.0 * markerMaf;

   for(int g=0;g<trans.persons;g++)
   {
      if(genotype[g]==-1.0)
	 genotype[g] = 0.0;
      else
	 genotype[g] -= mean;
   }

   return status;
}

int PreMeta::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,int markerid,FILE * log)
{
   //status=0: polymorphic
   //status=1: monomorphic or allele count greater than 2 
   //status=2: marker name from DAT file is incorrect and skip this marker
   int status = 1;

   StringArray markerName;
   markerName.AddTokens(ped.markerNames[markerid],":");
   if(markerName.Length()<=1)
   {
      printf("WARNING: marker name %s is not in CHR:POS format; this marker will be skipped from analysis.\n",ped.markerNames[markerid].c_str());
      fprintf(log,"WARNING: marker name %s is not in CHR:POS format; this marker will be skipped from analysis.\n",ped.markerNames[markerid].c_str());
      status=2;
      return status;
   }

   chr = markerName[0];
   pos = atoi(markerName[1].c_str());
markerName.Clear();

   founderMaf = markerMaf = 0.0;
   refAllele = rareAllele=".";

   genotype.Dimension(trans.persons);
   genotype.Zero();
   het=0;hom1=0;hom2=0;
   int numMiss=0;
   double mean =0.0;
   int n=0,nf=0;
   double maf=0.0,fmaf=0.0;
   callRate = 0.0;
   hwe_pvalue = 0.0;
   mac =0.0;

   int alleleCount = CountAlleles(ped, markerid);

   if(alleleCount==1)
   {
      rareAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
      refAllele = rareAllele;
      return status;
   }

   if(alleleCount!=2)
      return status;

   refAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
   rareAllele = ped.GetMarkerInfo(markerid)->alleleLabels[2];

   int idx=0;
   bool XStatus = false;
   if(chr==xLabel && pos>=Xstart && pos<=Xend)
      XStatus=true;

   for(int f=0;f<ped.familyCount;f++)
   {
      for(int p=0;p<trans.pPheno[f].Length();p++)
      {
	 if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	 {
	    numMiss++;
	    genotype[idx]=-1.0;
	    idx++;
	    continue;
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	 {
	    if(XStatus && ped[trans.pPheno[f][p]].sex==maleLabel)
	    {
	       hom2++;
	       genotype[idx]=2.0;
	    }
	    else
	    {
	       het++;
	       genotype[idx]=1.0 ;
	    }
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	 {
	    if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	    {
	       hom2++;
	       genotype[idx]=2.0;
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	    {
	       hom1++;
	       genotype[idx]=0.0;
	    }
	 }
	 n++;
	 maf+=genotype[idx];
	 int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	 if(founder != -1)
	 {
	    fmaf += genotype[idx];
	    nf++;
	 }
	 idx++;
      }
   }

   mac = maf;
   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0.0)
      fmaf = 1.0/(2.0*nf);

   callRate = 1.0-(double) numMiss/trans.persons;
   hwe_pvalue=0.0;
   //Get hwe_pvalue using the entire sample
   if(het==0 && hom1==0 && hom2==0)
   {
      //printf("marker %s:%d has problem.\n",chr.c_str(),pos);
      refAllele = rareAllele = ".";
      status=1;
      return status;
   }
   hwe_pvalue = SNPHWE(het,hom1,hom2);

   founderMaf = fmaf;
   markerMaf = maf;

   if(maf==0.0 || 1.0-maf==0.0)
      return status;

   status = 0;

   /*
      if(markerMaf > 0.5)
      {
      String tmp;
      tmp=refAllele;
      refAllele = rareAllele;
      rareAllele = tmp;
      markerMaf = 1.0-markerMaf;
      founderMaf = 1.0-founderMaf;
      for(int g=0;g<trans.persons;g++)
      {
      if(genotype[g]!=-1.0)
      genotype[g] = 2.0 -genotype[g];
      }
      int tmp_hom2 = hom2;
      hom2 = hom1;
      hom1 = tmp_hom2;
      hwe_pvalue = SNPHWE(het,hom1,hom2);
      }
    */
   mean = 2.0 * markerMaf;

   for(int g=0;g<trans.persons;g++)
   {
      if(genotype[g]==-1.0)
	 genotype[g] = 0.0;
      else
	 genotype[g] -= mean;
   }

   return status;
}

void PreMeta::PrintHeader(IFILE output,FastFit & engine,FastTransform & trans,Pedigree & ped)
{
   double sigma_g2 = engine.sigma_g2Hat;
   double h2 = 100.0*sigma_g2/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv =100.0*engine.sigma_s2Hat/trans.traitVar;
   //printf("simga_g2 is: %g,sigma_e2 is: %g,betaHat is: %g\n",sigma_g2,engine.sigma_e2Hat,engine.betaHat[0]);

   /*
      if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
      {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
      }
      else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
      {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
      }
      else if(AutoFit::fitSharedEnvironment)
      {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
      }
      else if(AutoFit::fitX)
      {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
      }
      fprintf(output,"===================================================================\n");
    */
   //fprintf(output,"\nCHR\tPOS\t#FOUNDERS\tFOUNDER_AF\tALL_AF\tINFORMATIVE_AC\tREF_ALLELE\tALT_ALLELE\tALT_ALLELE_EFFSIZE\tPVALUE\n");
   ifprintf(output,"##Heritability=%.3f%\n",h2);
   if(AutoFit::fitSharedEnvironment)
      ifprintf(output,"##SharedEnv=%.3f%\n",sharedEnv);
   if(AutoFit::fitX)
      ifprintf(output,"##Xheritability=%.3f%\n",h2X);
   if(stdErr)
      ifprintf(output,"#CHROM\tPOS\tREF\tALT\tN_INFORMATIVE\tFOUNDER_AF\tALL_AF\tINFORMATIVE_ALT_AC\tCALL_RATE\tHWE_PVALUE\tN_REF\tN_HET\tN_ALT\tU_STAT\tSQRT_V_STAT\tALT_EFFSIZE\tPVALUE\tSE\n");
   else
      ifprintf(output,"#CHROM\tPOS\tREF\tALT\tN_INFORMATIVE\tFOUNDER_AF\tALL_AF\tINFORMATIVE_ALT_AC\tCALL_RATE\tHWE_PVALUE\tN_REF\tN_HET\tN_ALT\tU_STAT\tSQRT_V_STAT\tALT_EFFSIZE\tPVALUE\n");
}

